import pkgutil
import csv

import numpy as np


def avail_approaches(pkg):
    """Create list of available modules.

    Parameters
    ----------
    pkg : module
        module to inspect

    Returns
    -------
    method : list
        A list of available submodules
    """
    methods = [
        modname
        for importer, modname, ispkg in pkgutil.walk_packages(path=pkg.__path__)
        if modname not in ["common_args", "directions", "sobol_sequence"]
        and "test" not in modname
    ]

    return methods


def read_param_file(filename, delimiter=None):
    """Unpacks a parameter file into a dictionary

    Reads a parameter file of format::

        Param1,0,1,Group1,dist1
        Param2,0,1,Group2,dist2
        Param3,0,1,Group3,dist3

    (Group and Dist columns are optional)

    Returns a dictionary containing:
        - names - the names of the parameters
        - bounds - a list of lists of lower and upper bounds
        - num_vars - a scalar indicating the number of variables
                     (the length of names)
        - groups - a list of group names (strings) for each variable
        - dists - a list of distributions for the problem,
                    None if not specified or all uniform

    Parameters
    ----------
    filename : str
        The path to the parameter file
    delimiter : str, default=None
        The delimiter used in the file to distinguish between columns

    """
    names = []
    bounds = []
    groups = []
    dists = []
    num_vars = 0
    fieldnames = ["name", "lower_bound", "upper_bound", "group", "dist"]

    with open(filename, "r") as csvfile:
        dialect = csv.Sniffer().sniff(csvfile.read(1024), delimiters=delimiter)
        csvfile.seek(0)
        reader = csv.DictReader(csvfile, fieldnames=fieldnames, dialect=dialect)
        for row in reader:
            if row["name"].strip().startswith("#"):
                pass
            else:
                num_vars += 1
                names.append(row["name"])
                bounds.append([float(row["lower_bound"]), float(row["upper_bound"])])

                # If the fourth column does not contain a group name, use
                # the parameter name
                if row["group"] is None:
                    groups.append(row["name"])
                elif row["group"] == "NA":
                    groups.append(row["name"])
                else:
                    groups.append(row["group"])

                # If the fifth column does not contain a distribution
                # use uniform
                if row["dist"] is None:
                    dists.append("unif")
                else:
                    dists.append(row["dist"])

    if groups == names:
        groups = None
    elif len(set(groups)) == 1:
        raise ValueError(
            """Only one group defined, results will not be
            meaningful"""
        )

    # setting dists to none if all are uniform
    # because non-uniform scaling is not needed
    if all([d == "unif" for d in dists]):
        dists = None

    return {
        "names": names,
        "bounds": bounds,
        "num_vars": num_vars,
        "groups": groups,
        "dists": dists,
    }


def _check_groups(problem):
    """Check if there is more than 1 group."""
    groups = problem.get("groups")
    if not groups:
        return False

    if len(set(groups)) == 1:
        return False
    return groups


def _check_bounds(bounds):
    """Check user supplied distribution bounds for validity.

    Parameters
    ----------
    problem : dict
        The problem definition

    Returns
    -------
    tuple : containing upper and lower bounds
    """
    b = np.array(bounds)

    lower_bounds = b[:, 0]
    upper_bounds = b[:, 1]

    if np.any(lower_bounds >= upper_bounds):
        raise ValueError("Bounds are not legal")

    return lower_bounds, upper_bounds



__all__ = ["plot"]

# magic string indicating DF columns holding conf bound values
CONF_COLUMN = "_conf"


def plot(Si_df, ax=None):
    """Create bar chart of results.


    Examples
    --------
        >>> from SALib.plotting.bar import plot as barplot
        >>> from SALib.test_functions import Ishigami
        >>>
        >>> # See README for example problem specification
        >>>
        >>> X = saltelli.sample(problem, 512)
        >>> Y = Ishigami.evaluate(X)
        >>> Si = sobol.analyze(problem, Y, print_to_console=False)
        >>> total, first, second = Si.to_df()
        >>> barplot(total)


    Parameters
    ----------
    * Si_df: pd.DataFrame, of sensitivity results

    Returns
    -------
    * ax : matplotlib axes object
    """
    conf_cols = Si_df.columns.str.contains(CONF_COLUMN)

    confs = Si_df.loc[:, conf_cols]
    confs.columns = [c.replace(CONF_COLUMN, "") for c in confs.columns]

    Sis = Si_df.loc[:, ~conf_cols]

    ax = Sis.plot(kind="bar", yerr=confs, ax=ax)
    return ax



from typing import Dict
import math
import warnings

import numpy as np

from . import common_args
from . import sobol_sequence
from ..util import scale_samples, read_param_file, compute_groups_matrix, _check_groups


def sample(
    problem: Dict, N: int, calc_second_order: bool = True, skip_values: int = None
):
    """Generates model inputs using Saltelli's extension of the Sobol' sequence

    The Sobol' sequence is a popular quasi-random low-discrepancy sequence used
    to generate uniform samples of parameter space.

    Returns a NumPy matrix containing the model inputs using Saltelli's
    sampling scheme.
    Saltelli's scheme extends the Sobol' sequence in a way to reduce
    the error rates in the resulting sensitivity index calculations. If
    `calc_second_order` is False, the resulting matrix has ``N * (D + 2)``
    rows, where ``D`` is the number of parameters.
    If `calc_second_order` is `True`, the resulting matrix has ``N * (2D + 2)``
    rows.
    These model inputs are intended to be used with
    :func:`SALib.analyze.sobol.analyze`.

    .. deprecated:: 1.4.6

    Notes
    -----
    The initial points of the Sobol' sequence has some repetition (see Table 2
    in Campolongo [1]), which can be avoided by setting the `skip_values`
    parameter. Skipping values reportedly improves the uniformity of samples.
    It has been shown that naively skipping values may reduce accuracy,
    increasing the number of samples needed to achieve convergence
    (see Owen [2]).

    A recommendation adopted here is that both `skip_values` and `N` be a power
    of 2, where `N` is the desired number of samples (see [2] and discussion in
    [5] for further context). It is also suggested therein that
    ``skip_values >= N``.

    The method now defaults to setting `skip_values` to a power of two that is
    ``>= N``. If `skip_values` is provided, the method now raises a UserWarning
    in cases where sample sizes may be sub-optimal according to the
    recommendation above.

    Parameters
    ----------
    problem : dict
        The problem definition
    N : int
        The number of samples to generate.
        Ideally a power of 2 and <= `skip_values`.
    calc_second_order : bool
        Calculate second-order sensitivities (default True)
    skip_values : int or None
        Number of points in Sobol' sequence to skip, ideally a value of base 2
        (default: a power of 2 >= N, or 16; whichever is greater)


    References
    ----------
    1. Campolongo, F., Saltelli, A., Cariboni, J., 2011.
           From screening to quantitative sensitivity analysis.
           A unified approach.
           Computer Physics Communications 182, 978-988.
           https://doi.org/10.1016/j.cpc.2010.12.039

    2. Owen, A. B., 2020.
           On dropping the first Sobol' point.
           arXiv:2008.08051 [cs, math, stat].
           Available at: http://arxiv.org/abs/2008.08051
           (Accessed: 20 April 2021).

    3. Saltelli, A., 2002.
           Making best use of model evaluations to compute sensitivity indices.
           Computer Physics Communications 145, 280-297.
           https://doi.org/10.1016/S0010-4655(02)00280-1

    4. Sobol', I.M., 2001.
           Global sensitivity indices for nonlinear mathematical models and
           their Monte Carlo estimates.
           Mathematics and Computers in Simulation,
           The Second IMACS Seminar on Monte Carlo Methods 55, 271-280.
           https://doi.org/10.1016/S0378-4754(00)00270-6

    5. Discussion: https://github.com/scipy/scipy/pull/10844
           https://github.com/scipy/scipy/pull/10844#issuecomment-672186615
           https://github.com/scipy/scipy/pull/10844#issuecomment-673029539
    """
    warnings.warn(
        "`salib.sample.saltelli` will be removed in SALib 1.5. Please use"
        " `salib.sample.sobol`",
        category=DeprecationWarning,
        stacklevel=2,
    )

    # bit-shift test to check if `N` == 2**n
    if not ((N & (N - 1) == 0) and (N != 0 and N - 1 != 0)):
        msg = f"""
        Convergence properties of the Sobol' sequence is only valid if
        `N` ({N}) is equal to `2^n`.
        """
        warnings.warn(msg)

    if skip_values is None:
        # If not specified, set skip_values to next largest power of 2
        skip_values = int(2 ** math.ceil(math.log(N) / math.log(2)))

        # 16 is arbitrarily selected here to avoid initial points
        # for very low sample sizes
        skip_values = max(skip_values, 16)

    elif skip_values > 0:
        M = skip_values
        if not ((M & (M - 1) == 0) and (M != 0 and M - 1 != 0)):
            msg = f"""
            Convergence properties of the Sobol' sequence is only valid if
            `skip_values` ({M}) is a power of 2.
            """
            warnings.warn(msg)

        # warning when N > skip_values
        # see https://github.com/scipy/scipy/pull/10844#issuecomment-673029539
        n_exp = int(math.log(N, 2))
        m_exp = int(math.log(M, 2))
        if n_exp > m_exp:
            msg = (
                "Convergence may not be valid as the number of "
                "requested samples is"
                f" > `skip_values` ({N} > {M})."
            )
            warnings.warn(msg)
    elif skip_values == 0:
        warnings.warn("Duplicate samples will be taken as no points are skipped.")
    else:
        assert (
            isinstance(skip_values, int) and skip_values >= 0
        ), "`skip_values` must be a positive integer."

    D = problem["num_vars"]
    groups = _check_groups(problem)

    if not groups:
        Dg = problem["num_vars"]
    else:
        G, group_names = compute_groups_matrix(groups)
        Dg = len(set(group_names))

    # Create base sequence - could be any type of sampling
    base_sequence = sobol_sequence.sample(N + skip_values, 2 * D)

    if calc_second_order:
        saltelli_sequence = np.zeros([(2 * Dg + 2) * N, D])
    else:
        saltelli_sequence = np.zeros([(Dg + 2) * N, D])
    index = 0

    for i in range(skip_values, N + skip_values):

        # Copy matrix "A"
        for j in range(D):
            saltelli_sequence[index, j] = base_sequence[i, j]

        index += 1

        # Cross-sample elements of "B" into "A"
        for k in range(Dg):
            for j in range(D):
                if (not groups and j == k) or (groups and group_names[k] == groups[j]):
                    saltelli_sequence[index, j] = base_sequence[i, j + D]
                else:
                    saltelli_sequence[index, j] = base_sequence[i, j]

            index += 1

        # Cross-sample elements of "A" into "B"
        # Only needed if you're doing second-order indices (true by default)
        if calc_second_order:
            for k in range(Dg):
                for j in range(D):
                    if (not groups and j == k) or (
                        groups and group_names[k] == groups[j]
                    ):
                        saltelli_sequence[index, j] = base_sequence[i, j]
                    else:
                        saltelli_sequence[index, j] = base_sequence[i, j + D]

                index += 1

        # Copy matrix "B"
        for j in range(D):
            saltelli_sequence[index, j] = base_sequence[i, j + D]

        index += 1

    saltelli_sequence = scale_samples(saltelli_sequence, problem)
    return saltelli_sequence


def cli_parse(parser):
    """Add method specific options to CLI parser.

    Parameters
    ----------
    parser : argparse object

    Returns
    -------
    Updated argparse object
    """
    parser.add_argument(
        "--max-order",
        type=int,
        required=False,
        default=2,
        choices=[1, 2],
        help="Maximum order of sensitivity indices \
                           to calculate",
    )
    parser.add_argument(
        "--skip-values",
        type=int,
        required=False,
        default=None,
        help="Number of sample points to skip \
                            (default: next largest power of 2 from `samples`)",
    )

    # hacky way to remove an argument (seed option not relevant for Saltelli)
    remove_opts = [x for x in parser._actions if x.dest == "seed"]
    [
        parser._handle_conflict_resolve(None, [("--seed", x), ("-s", x)])
        for x in remove_opts
    ]

    return parser


def cli_action(args):
    """Run sampling method

    Parameters
    ----------
    args : argparse namespace
    """
    problem = read_param_file(args.paramfile)
    param_values = sample(
        problem,
        args.samples,
        calc_second_order=(args.max_order == 2),
        skip_values=args.skip_values,
    )
    np.savetxt(
        args.output,
        param_values,
        delimiter=args.delimiter,
        fmt="%." + str(args.precision) + "e",
    )


if __name__ == "__main__":
    common_args.run_cli(cli_parse, cli_action)



import warnings
import importlib
from types import MethodType
from functools import partial, wraps

from multiprocess import Pool, cpu_count
import numpy as np

import SALib.sample as samplers
import SALib.analyze as analyzers

from SALib.util import avail_approaches
from SALib.util.results import ResultDict


ptqdm_available = True
try:
    from p_tqdm import p_imap
except ImportError:
    ptqdm_available = False

try:
    from pathos.pp import ParallelPythonPool as pp_Pool

    pathos_available = True
except ImportError:
    pathos_available = False


__all__ = ["ProblemSpec"]


class ProblemSpec(dict):
    """Dictionary-like object representing an SALib Problem specification.

    Attributes
    ----------
    samples : np.array, of generated samples
    results : np.array, of evaluations (i.e., model outputs)
    analysis : np.array or dict, of sensitivity indices
    """

    def __init__(self, *args, **kwargs):
        super(ProblemSpec, self).__init__(*args, **kwargs)

        _check_spec_attributes(self)

        self._samples = None
        self._results = None
        self._analysis = None

        self["num_vars"] = len(self["names"])
        if "groups" not in self:
            self["groups"] = None

        self._add_samplers()
        self._add_analyzers()

    @property
    def samples(self):
        return self._samples

    @samples.setter
    def samples(self, vals):
        cols = vals.shape[1]
        if cols != self["num_vars"]:
            msg = "Mismatched sample size: Expected "
            msg += "{} cols, got {}".format(self["num_vars"], cols)
            raise ValueError(msg)

        self._samples = vals

        # Clear results to avoid confusion
        self._results = None

    @property
    def results(self):
        return self._results

    @results.setter
    def results(self, vals):
        val_shape = vals.shape
        if len(val_shape) == 1:
            cols = 1
        else:
            cols = vals.shape[1]

        out_cols = self.get("outputs", None)
        if out_cols is None:
            if cols == 1:
                self["outputs"] = ["Y"]
            else:
                self["outputs"] = [f"Y{i}" for i in range(1, cols + 1)]
        else:
            if cols != len(self["outputs"]):
                msg = "Mismatched sample size: Expected "
                msg += "{} cols, got {}".format(self["outputs"], cols)
                raise ValueError(msg)

        self._results = vals

    @property
    def analysis(self):
        return self._analysis

    def sample(self, func, *args, **kwargs):
        """Create sample using given function.

        Parameters
        ----------
        func : function,
            Sampling method to use. The given function must accept the SALib
            problem specification as the first parameter and return a numpy
            array.

        *args : list,
            Additional arguments to be passed to `func`

        **kwargs : dict,
            Additional keyword arguments passed to `func`

        Returns
        -------
        self : ProblemSpec object
        """
        # Clear model output and analysis results to avoid confusion
        # especially if samples are forcibly changed...
        self._analysis = None
        self._results = None
        self._samples = func(self, *args, **kwargs)

        return self

    def set_samples(self, samples: np.ndarray):
        """Set previous samples used."""
        self.samples = samples

        return self

    def set_results(self, results: np.ndarray):
        """Set previously available model results."""
        if self.samples is not None:
            assert self.samples.shape[0] == results.shape[0], (
                "Provided result array does not match existing number of"
                " existing samples!"
            )

        self.results = results

        return self

    def evaluate(self, func, *args, **kwargs):
        """Evaluate a given model.

        Parameters
        ----------
        func : function,
            Model, or function that wraps a model, to be run/evaluated.
            The provided function is required to accept a numpy array of
            inputs as its first parameter and must return a numpy array of
            results.

        *args : list,
            Additional arguments to be passed to `func`

        nprocs : int,
            If specified, attempts to parallelize model evaluations

        **kwargs : dict,
            Additional keyword arguments passed to `func`

        Returns
        -------
        self : ProblemSpec object
        """
        if "nprocs" in kwargs:
            nprocs = kwargs.pop("nprocs")
            return self.evaluate_parallel(func, *args, nprocs=nprocs, **kwargs)

        self.results = func(self._samples, *args, **kwargs)

        return self

    def evaluate_parallel(self, func, *args, nprocs=None, **kwargs):
        """Evaluate model locally in parallel.

        All detected processors will be used if `nprocs` is not specified.

        Parameters
        ----------
        func : function,
            Model, or function that wraps a model, to be run in parallel.
            The provided function needs to accept a numpy array of inputs as
            its first parameter and must return a numpy array of results.

        nprocs : int,
            Number of processors to use.
            Capped to the number of available processors.

        *args : list,
            Additional arguments to be passed to `func`

        **kwargs : dict,
            Additional keyword arguments passed to `func`

        Returns
        -------
        self : ProblemSpec object
        """
        warnings.warn(
            "Parallel evaluation is an experimental feature and may not work."
        )

        if self._samples is None:
            raise RuntimeError("Sampling not yet conducted")

        max_procs = cpu_count()
        if nprocs is None:
            nprocs = max_procs
        else:
            if nprocs > max_procs:
                warnings.warn(
                    f"{nprocs} processors requested but only {max_procs} found."
                )
            nprocs = min(max_procs, nprocs)

        # Create wrapped partial function to allow passing of additional args
        tmp_f = self._wrap_func(func, *args, **kwargs)

        # Split into even chunks
        chunks = np.array_split(self._samples, int(nprocs), axis=0)

        if ptqdm_available:
            # Display progress bar if available
            res = p_imap(tmp_f, chunks, num_cpus=nprocs)
        else:
            with Pool(nprocs) as pool:
                res = list(pool.imap(tmp_f, chunks))

        self.results = self._collect_results(res)

        return self

    def evaluate_distributed(
        self, func, *args, nprocs=1, servers=None, verbose=False, **kwargs
    ):
        """Distribute model evaluation across a cluster.

        Usage Conditions:

        * The provided function needs to accept a numpy array of inputs as
          its first parameter
        * The provided function must return a numpy array of results


        Parameters
        ----------
        func : function,
            Model, or function that wraps a model, to be run in parallel

        nprocs : int,
            Number of processors to use for each node. Defaults to 1.

        servers : list[str] or None,
            IP addresses or alias for each server/node to use.

        verbose : bool,
            Display job execution statistics. Defaults to False.

        *args : list,
            Additional arguments to be passed to `func`

        **kwargs : dict,
            Additional keyword arguments passed to `func`

        Returns
        -------
        self : ProblemSpec object
        """
        if not pathos_available:
            raise RuntimeError(
                "Pathos is required to run in distributed mode. Please install"
                " with `pip install pathos` or"
                " `conda install pathos -c conda-forge`"
            )

        if verbose:
            from pathos.parallel import stats

        warnings.warn(
            "Distributed evaluation is an untested experimental feature and"
            " may not work."
        )

        workers = pp_Pool(nprocs, servers=servers)

        # Split into even chunks
        chunks = np.array_split(self._samples, int(nprocs) * len(servers), axis=0)

        tmp_f = self._wrap_func(func)

        res = list(workers.map(tmp_f, chunks))

        self.results = self._collect_results(res)

        if verbose:
            print(stats(), "\n")

        workers.clear()

        return self

    def analyze(self, func, *args, **kwargs):
        """Analyze sampled results using given function.

        Parameters
        ----------
        func : function,
            Analysis method to use. The provided function must accept the
            problem specification as the first parameter, X values if needed,
            Y values, and return a numpy array.

        *args : list,
            Additional arguments to be passed to `func`

        nprocs : int,
            If specified, attempts to parallelize model evaluations

        **kwargs : dict,
            Additional keyword arguments passed to `func`

        Returns
        -------
        self : ProblemSpec object
        """
        if self["num_vars"] == 1 or (self["groups"] and len("groups") == 1):
            msg = (
                "There is only a single parameter or group defined. There is "
                "no point in conducting sensitivity analysis as any and all"
                "effect(s) will be mapped to the single parameter/group."
            )
            raise ValueError(msg)

        if "nprocs" in kwargs:
            # Call parallel method instead
            return self.analyze_parallel(func, *args, **kwargs)

        if self._results is None:
            raise RuntimeError("Model not yet evaluated")

        if "X" in func.__code__.co_varnames:
            # enforce passing of X if expected
            func = partial(func, *args, X=self._samples, **kwargs)
        else:
            func = partial(func, *args, **kwargs)

        out_cols = self.get("outputs", None)
        if out_cols is None:
            if len(self._results.shape) == 1:
                self["outputs"] = ["Y"]
            else:
                num_cols = self._results.shape[1]
                self["outputs"] = [f"Y{i}" for i in range(1, num_cols + 1)]

        if len(self["outputs"]) > 1:
            self._analysis = {}
            for i, out in enumerate(self["outputs"]):
                self._analysis[out] = func(self, Y=self._results[:, i])
        else:
            self._analysis = func(self, Y=self._results)

        return self

    def analyze_parallel(self, func, *args, nprocs=None, **kwargs):
        """Analyze sampled results using the given function in parallel.

        Parameters
        ----------
        func : function,
            Analysis method to use. The provided function must accept the
            problem specification as the first parameter, X values if needed,
            Y values, and return a numpy array.

        *args : list,
            Additional arguments to be passed to `func`

        nprocs : int,
            Number of processors to use.
            Capped to the number of outputs or available processors.

        **kwargs : dict,
            Additional keyword arguments passed to `func`

        Returns
        -------
        self : ProblemSpec object
        """
        warnings.warn("Parallel analysis is an experimental feature and may not work.")

        if self._results is None:
            raise RuntimeError("Model not yet evaluated")

        if "X" in func.__code__.co_varnames:
            # enforce passing of X if expected
            func = partial(func, *args, X=self._samples, **kwargs)
        else:
            func = partial(func, *args, **kwargs)

        out_cols = self.get("outputs", None)
        if out_cols is None:
            if len(self._results.shape) == 1:
                self["outputs"] = ["Y"]
            else:
                num_cols = self._results.shape[1]
                self["outputs"] = [f"Y{i}" for i in range(1, num_cols + 1)]

        # Cap number of processors used
        Yn = len(self["outputs"])
        if Yn == 1:
            # Only single output, cannot parallelize
            warnings.warn(
                f"Analysis was not parallelized: {nprocs} processors requested"
                f" for 1 output."
            )

            res = func(self, Y=self._results)
        else:
            max_procs = cpu_count()
            if nprocs is None:
                nprocs = max_procs
            else:
                nprocs = min(Yn, nprocs, max_procs)

            if ptqdm_available:
                # Display progress bar if available
                res = p_imap(
                    lambda y: func(self, Y=y),
                    [self._results[:, i] for i in range(Yn)],
                    num_cpus=nprocs,
                )
            else:
                with Pool(nprocs) as pool:
                    res = list(
                        pool.imap(
                            lambda y: func(self, Y=y),
                            [self._results[:, i] for i in range(Yn)],
                        )
                    )

        # Assign by output name if more than 1 output, otherwise
        # attach directly
        if Yn > 1:
            self._analysis = {}
            for out, Si in zip(self["outputs"], list(res)):
                self._analysis[out] = Si
        else:
            self._analysis = res

        return self

    def to_df(self):
        """Convert results to Pandas DataFrame."""
        an_res = self._analysis
        if isinstance(an_res, ResultDict):
            return an_res.to_df()
        elif isinstance(an_res, dict):
            # case where analysis result is a dict of ResultDicts
            return [an.to_df() for an in list(an_res.values())]

        raise RuntimeError("Analysis not yet conducted")

    def plot(self):
        """Plot results as a bar chart.

        Returns
        -------
        axes : matplotlib axes object
        """
        if self._analysis is None:
            raise RuntimeError("Analysis not yet conducted")

        num_rows = len(self["outputs"])
        if num_rows == 1:
            return self._analysis.plot()

        try:
            plt
        except NameError:
            import matplotlib.pyplot as plt

        num_cols = 1
        fk = list(self._analysis.keys())[0]
        if isinstance(self._analysis[fk].to_df(), (list, tuple)):
            # have to divide by 2 to account for CI columns
            num_cols = len(self._analysis[fk]) // 2

        p_width = max(num_cols * 3, 5)
        p_height = max(num_rows * 3, 6)
        _, axes = plt.subplots(
            num_rows, num_cols, sharey=True, figsize=(p_width, p_height)
        )
        for res, ax in zip(self._analysis, axes):
            self._analysis[res].plot(ax=ax)

            try:
                ax[0].set_title(res)
            except TypeError:
                ax.set_title(res)

        plt.tight_layout()

        return axes

    def heatmap(
        self, metric: str = None, index: str = None, title: str = None, ax=None
    ):
        """Plot results as a heatmap.

        Parameters
        ----------
        metric : str or None, name of output to analyze (display all if `None`)
        index : str or None, name of index to plot, dependent on what
                    analysis was conducted (ST, S1, etc; displays all if `None`)
        title : str, title of plot to use (defaults to the same as `metric`)
        ax : axes object, matplotlib axes object to use for plot.
                Creates a new figure if not provided.

        Returns
        -------
        ax : matplotlib axes object
        """
        from SALib.plotting.heatmap import heatmap  # type: ignore

        return heatmap(self, metric, index, title, ax)

    def _wrap_func(self, func, *args, **kwargs):
        # Create wrapped partial function to allow passing of additional args
        tmp_f = func
        if (len(args) > 0) or (len(kwargs) > 0):
            tmp_f = lambda x: func(x, *args, **kwargs)  # noqa

        return tmp_f

    def _collect_results(self, res):
        res_shape = res[0].shape
        if len(res_shape) > 1:
            res_shape = (len(self._samples), *res_shape[1:])
        else:
            res_shape = len(self._samples)

        final_res = np.empty(res_shape)

        # Collect results
        # Cannot enumerate over this as the length
        # of individual results may vary
        i = 0
        for r in res:
            r_len = len(r)
            final_res[i : i + r_len] = r
            i += r_len

        return final_res

    def _method_creator(self, func, method):
        """Generate convenience methods for specified `method`."""

        @wraps(func)
        def modfunc(self, *args, **kwargs):
            return getattr(self, method)(func, *args, **kwargs)

        return modfunc

    def _add_samplers(self):
        """Dynamically add available SALib samplers as ProblemSpec methods."""
        for sampler in avail_approaches(samplers):
            func = getattr(
                importlib.import_module("SALib.sample.{}".format(sampler)), "sample"
            )
            method_name = "sample_{}".format(sampler.replace("_sampler", ""))

            self.__setattr__(
                method_name, MethodType(self._method_creator(func, "sample"), self)
            )

    def _add_analyzers(self):
        """Dynamically add available SALib analyzers as ProblemSpec methods."""
        for analyzer in avail_approaches(analyzers):
            func = getattr(
                importlib.import_module("SALib.analyze.{}".format(analyzer)), "analyze"
            )
            method_name = "analyze_{}".format(analyzer.replace("_analyzer", ""))

            self.__setattr__(
                method_name, MethodType(self._method_creator(func, "analyze"), self)
            )

    def __str__(self):
        if self._samples is not None:
            arr_shape = self._samples.shape
            if len(arr_shape) == 1:
                arr_shape = (arr_shape[0], 1)
            nr, nx = arr_shape
            print("Samples:")
            print(f"\t{nx} parameters:", self["names"])
            print(f"\t{nr} evaluations", "\n")
        if self._results is not None:
            arr_shape = self._results.shape
            if len(arr_shape) == 1:
                arr_shape = (arr_shape[0], 1)
            nr, ny = arr_shape
            print("Outputs:")
            print(f"\t{ny} outputs:", self["outputs"])
            print(f"\t{nr} evaluations", "\n")
        if self._analysis is not None:
            print("Analysis:")
            an_res = self._analysis

            allowed_types = (list, tuple)
            if isinstance(an_res, ResultDict):
                an_res = an_res.to_df()
                if not isinstance(an_res, allowed_types):
                    print(an_res, "\n")
                else:
                    for df in an_res:
                        print(df, "\n")
            elif isinstance(an_res, dict):
                for res_name in an_res:
                    print("{}:".format(res_name))
                    dfs = an_res[res_name].to_df()
                    if isinstance(dfs, allowed_types):
                        for df in dfs:
                            print(df, "\n")
                    else:
                        print(dfs, "\n")
        return ""


def _check_spec_attributes(spec: ProblemSpec):
    assert "names" in spec, "Names not defined"
    assert "bounds" in spec, "Bounds not defined"
    assert len(spec["bounds"]) == len(
        spec["names"]
    ), f"""Number of bounds do not match number of names
        Number of names:
        {len(spec['names'])} | {spec['names']}
        ----------------
        Number of bounds: {len(spec['bounds'])}
        """



"""A set of utility functions

"""
from collections import OrderedDict
from typing import Dict, Tuple
import warnings

import numpy as np  # type: ignore
import scipy as sp  # type: ignore
from typing import List

from .util_funcs import (  # noqa: F401, E402
    avail_approaches,
    read_param_file,
    _check_bounds,
    _check_groups,
)  # noqa: F401, E402
from .problem import ProblemSpec  # noqa: F401, E402
from .results import ResultDict  # noqa: F401, E402


__all__ = ["scale_samples", "read_param_file", "avail_approaches"]


def _scale_samples(params: np.ndarray, bounds: List):
    """Rescale samples in 0-to-1 range to arbitrary bounds

    Parameters
    ----------
    params : numpy.ndarray
        numpy array of dimensions `num_params`-by-:math:`N`,
        where :math:`N` is the number of samples

    bounds : list
        list of lists of dimensions `num_params`-by-2
    """
    # Check bounds are legal (upper bound is greater than lower bound)
    lower_bounds, upper_bounds = _check_bounds(bounds)

    if np.any(lower_bounds >= upper_bounds):
        raise ValueError(
            "Bounds are not legal (upper bound must be greater than lower bound)"
        )

    # This scales the samples in-place, by using the optional output
    # argument for the numpy ufunctions
    # The calculation is equivalent to:
    #   sample * (upper_bound - lower_bound) + lower_bound
    np.add(
        np.multiply(params, (upper_bounds - lower_bounds), out=params),
        lower_bounds,
        out=params,
    )


def scale_samples(params: np.ndarray, problem: Dict):
    """Scale samples based on specified distribution (defaulting to uniform).

    Adds an entry to the problem specification to indicate samples have been
    scaled to maintain backwards compatibility (`sample_scaled`).

    Parameters
    ----------
    params : np.ndarray,
        numpy array of dimensions `num_params`-by-:math:`N`,
        where :math:`N` is the number of samples
    problem : dictionary,
        SALib problem specification

    Returns
    -------
    np.ndarray, scaled samples
    """
    bounds = problem["bounds"]
    dists = problem.get("dists")

    if dists is None:
        _scale_samples(params, bounds)
    else:
        if params.shape[1] != len(dists):
            msg = "Mismatch in number of parameters and distributions.\n"
            msg += "Num parameters: {}".format(params.shape[1])
            msg += "Num distributions: {}".format(len(dists))
            raise ValueError(msg)

        params = _nonuniform_scale_samples(params, bounds, dists)

    problem["sample_scaled"] = True

    return params
    # limited_params = limit_samples(params, upper_bound, lower_bound, dists)


def _unscale_samples(params, bounds):
    """Rescale samples from arbitrary bounds back to [0,1] range

    Parameters
    ----------
    bounds : list
        list of lists of dimensions num_params-by-2
    params : numpy.ndarray
        numpy array of dimensions num_params-by-N,
        where N is the number of samples
    """
    # Check bounds are legal (upper bound is greater than lower bound)
    b = np.array(bounds)
    lower_bounds = b[:, 0]
    upper_bounds = b[:, 1]

    if np.any(lower_bounds >= upper_bounds):
        raise ValueError(
            "Bounds are not legal (upper bound must be greater than lower bound)"
        )

    # This scales the samples in-place, by using the optional output
    # argument for the numpy ufunctions
    # The calculation is equivalent to:
    #   (sample - lower_bound) / (upper_bound - lower_bound)
    np.divide(
        np.subtract(params, lower_bounds, out=params),
        np.subtract(upper_bounds, lower_bounds),
        out=params,
    )


def _nonuniform_scale_samples(params, bounds, dists):
    """Rescale samples in 0-to-1 range to other distributions

    Parameters
    ----------
    params : numpy.ndarray
        numpy array of dimensions num_params-by-N,
        where N is the number of samples
    dists : list
        list of distributions, one for each parameter
            unif: uniform with lower and upper bounds
            triang: triangular with lower and upper bounds, as well as
                    location of peak
                    The location of peak is in percentage of width
                    e.g. :code:`[1.0, 3.0, 0.5]` indicates 1.0 to 3.0 with a
                    peak at 2.0

                    A soon-to-be deprecated two-value format assumes the lower
                    bound to be 0
                    e.g. :code:`[3, 0.5]` assumes 0 to 3, with a peak at 1.5
            norm: normal distribution with mean and standard deviation
            truncnorm: truncated normal distribution with upper and lower
                    bounds, mean and standard deviation
            lognorm: lognormal with ln-space mean and standard deviation
    """
    b = np.array(bounds, dtype=object)

    # initializing matrix for converted values
    conv_params = np.empty_like(params)

    # loop over the parameters
    for i in range(conv_params.shape[1]):
        # setting first and second arguments for distributions
        b1 = b[i][0]  # ending
        b2 = b[i][1]  # 0-1

        if dists[i] == "triang":
            if len(b[i]) == 3:
                loc_start = b[i][0]  # loc start
                b1 = b[i][1]  # triangular distribution end
                b2 = b[i][2]  # 0-1 aka c (the peak)
            elif len(b[i]) == 2:
                msg = (
                    "Two-value format for triangular distributions detected.\n"
                    "To remove this message, specify the distribution start, "
                    "end, and peak (three values) "
                    "instead of the current two-value format "
                    "(distribution end and peak, with start assumed to be 0)\n"
                    "The two-value format will be deprecated in SALib v1.5"
                )
                warnings.warn(msg, DeprecationWarning, stacklevel=2)

                loc_start = 0
                b1 = b[i][0]
                b2 = b[i][1]
            else:
                raise ValueError(
                    "Unknown triangular distribution specification. Check"
                    " problem specification."
                )

            # checking for correct parameters
            if b1 < 0 or b2 < 0 or b2 >= 1 or loc_start > b1:
                raise ValueError(
                    """Triangular distribution bound error: Scale must be
                    greater than zero; peak on interval [0,1], triangular
                    start value must be smaller than end value"""
                )
            else:
                conv_params[:, i] = sp.stats.triang.ppf(
                    params[:, i], c=b2, scale=b1 - loc_start, loc=loc_start
                )

        elif dists[i] == "unif":
            if b1 >= b2:
                raise ValueError(
                    """Uniform distribution: lower bound
                    must be less than upper bound"""
                )
            else:
                conv_params[:, i] = params[:, i] * (b2 - b1) + b1

        elif dists[i] == "norm":
            if b2 <= 0:
                raise ValueError("""Normal distribution: stdev must be > 0""")
            else:
                conv_params[:, i] = sp.stats.norm.ppf(params[:, i], loc=b1, scale=b2)

        # Truncated normal distribution
        # parameters are lower bound and upper bound, mean and stdev
        elif dists[i] == "truncnorm":
            b3 = b[i][2]
            b4 = b[i][3]
            if b4 <= 0:
                raise ValueError(
                    """Truncated normal distribution: stdev must
                    be > 0"""
                )
            if b1 >= b2:
                raise ValueError(
                    """Truncated normal distribution: lower bound
                    must be less than upper bound"""
                )
            else:
                conv_params[:, i] = sp.stats.truncnorm.ppf(
                    params[:, i], (b1 - b3) / b4, (b2 - b3) / b4, loc=b3, scale=b4
                )

        # lognormal distribution (ln-space, not base-10)
        # paramters are ln-space mean and standard deviation
        elif dists[i] == "lognorm":
            # checking for valid parameters
            if b2 <= 0:
                raise ValueError("""Lognormal distribution: stdev must be > 0""")
            else:
                conv_params[:, i] = np.exp(
                    sp.stats.norm.ppf(params[:, i], loc=b1, scale=b2)
                )

        else:
            valid_dists = ["unif", "triang", "norm", "truncnorm", "lognorm"]
            raise ValueError("Distributions: choose one of %s" % ", ".join(valid_dists))

    return conv_params


def extract_group_names(groups: List) -> Tuple:
    """Get a unique set of the group names.

    Reverts to parameter names (and number of parameters) if groups not
    defined.

    Parameters
    ----------
    groups : List


    Returns
    -------
    tuple : names, number of groups
    """
    names = list(OrderedDict.fromkeys(groups))
    number = len(names)

    return names, number


def compute_groups_matrix(groups: List):
    """Generate matrix which notes factor membership of groups

    Computes a k-by-g matrix which notes factor membership of groups
    where:
        k is the number of variables (factors)
        g is the number of groups
    Also returns a g-length list of unique group_names whose positions
    correspond to the order of groups in the k-by-g matrix

    Parameters
    ----------
    groups : List
        Group names corresponding to each variable

    Returns
    -------
    tuple
        containing group matrix assigning parameters to
        groups and a list of unique group names
    """
    num_vars = len(groups)
    unique_group_names, number_of_groups = extract_group_names(groups)

    indices = dict([(x, i) for (i, x) in enumerate(unique_group_names)])

    output = np.zeros((num_vars, number_of_groups), dtype=int)

    for parameter_row, group_membership in enumerate(groups):
        group_index = indices[group_membership]
        output[parameter_row, group_index] = 1

    return output, unique_group_names


def _define_problem_with_groups(problem: Dict) -> Dict:
    """
    Checks if the user defined the 'groups' key in the problem dictionary.
    If not, makes the 'groups' key equal to the variables names. In other
    words, the number of groups will be equal to the number of variables, which
    is equivalent to no groups.

    Parameters
    ----------
    problem : dict
        The problem definition

    Returns
    -------
    problem : dict
        The problem definition with the 'groups' key, even if the user doesn't
        define it
    """
    # Checks if there isn't a key 'groups' or if it exists and is set to 'None'
    if "groups" not in problem or not problem["groups"]:
        problem["groups"] = problem["names"]
    elif len(problem["groups"]) != problem["num_vars"]:
        raise ValueError(
            "Number of entries in 'groups' should be the same " "as in 'names'"
        )
    return problem


def _compute_delta(num_levels: int) -> float:
    """Computes the delta value from number of levels

    Parameters
    ---------
    num_levels : int
        The number of levels

    Returns
    -------
    float
    """
    return num_levels / (2.0 * (num_levels - 1))



import numpy as np
from typing import Dict

import numpy.random as rd
import warnings

from .local import LocalOptimisation
from .brute import BruteForce

from .strategy import SampleMorris

from SALib.sample import common_args
from SALib.util import (
    scale_samples,
    read_param_file,
    compute_groups_matrix,
    _define_problem_with_groups,
    _compute_delta,
    _check_groups,
)


__all__ = ["sample"]


def sample(
    problem: Dict,
    N: int,
    num_levels: int = 4,
    optimal_trajectories: int = None,
    local_optimization: bool = True,
    seed: int = None,
) -> np.ndarray:
    """Generate model inputs using the Method of Morris.

    Three variants of Morris' sampling for elementary effects is supported:

    - Vanilla Morris (see [1])
      when ``optimal_trajectories`` is ``None``/``False`` and
      ``local_optimization`` is ``False``
    - Optimised trajectories when ``optimal_trajectories=True`` using
        Campolongo's enhancements (see [2]) and optionally Ruano's enhancement
        (see [3]) when ``local_optimization=True``
    - Morris with groups when the problem definition specifies groups of
      parameters

    Results from these model inputs are intended to be used with
    :func:`SALib.analyze.morris.analyze`.

    Notes
    -----
    Campolongo et al., [2] introduces an optimal trajectories approach which
    attempts to maximize the parameter space scanned for a given number of
    trajectories (where `optimal_trajectories` :math:`\\in {2, ..., N}`).
    The approach accomplishes this aim by randomly generating a high number
    of possible trajectories (500 to 1000 in [2]) and selecting a subset of
    ``r`` trajectories which have the highest spread in parameter space.
    The ``r`` variable in [2] corresponds to the ``optimal_trajectories``
    parameter here.

    Calculating all possible combinations of trajectories can be
    computationally expensive. The number of factors makes little
    difference, but the ratio between number of optimal trajectories and the
    sample size results in an exponentially increasing number of scores that
    must be computed to find the optimal combination of trajectories. We
    suggest going no higher than 4 levels from a pool of 100 samples with this
    "brute force" approach.

    Ruano et al., [3] proposed an alternative approach with an iterative
    process that maximizes the distance between subgroups of generated
    trajectories, from which the final set of trajectories are selected, again
    maximizing the distance between each. The approach is not guaranteed to
    produce the most optimal spread of trajectories, but are at least locally
    maximized and significantly reduce the time taken to select trajectories.
    With ``local_optimization = True`` (which is default), it is possible to
    go higher than the previously suggested 4 levels from a pool of 100
    samples.

    Parameters
    ----------
    problem : dict
        The problem definition
    N : int
        The number of trajectories to generate
    num_levels : int, default=4
        The number of grid levels (should be even)
    optimal_trajectories : int
        The number of optimal trajectories to sample (between 2 and N)
    local_optimization : bool, default=True
        Flag whether to use local optimization according to Ruano et al. (2012)
        Speeds up the process tremendously for bigger N and num_levels.
        If set to ``False`` brute force method is used, unless ``gurobipy`` is
        available
    seed : int
        Seed to generate a random number

    Returns
    -------
    sample_morris : np.ndarray
        Array containing the model inputs required for Method of Morris.
        The resulting matrix has :math:`(G/D+1)*N/T` rows and :math:`D`
        columns, where :math:`D` is the number of parameters, :math:`G`
        is the number of groups (if no groups are selected, the number
        of parameters). :math:`T` is the number of trajectories
        :math:`N`, or `optimal_trajectories` if selected.

    References
    ----------
    1. Morris, M.D., 1991.
       Factorial Sampling Plans for Preliminary Computational Experiments.
       Technometrics 33, 161-174.
       https://doi.org/10.1080/00401706.1991.10484804

    2. Campolongo, F., Cariboni, J., & Saltelli, A. 2007.
       An effective screening design for sensitivity analysis of large
       models.
       Environmental Modelling & Software, 22(10), 1509-1518.
       https://doi.org/10.1016/j.envsoft.2006.10.004

    3. Ruano, M.V., Ribes, J., Seco, A., Ferrer, J., 2012.
       An improved sampling strategy based on trajectory design for
       application of the Morris method to systems with many input
       factors.
       Environmental Modelling & Software 37, 103-109.
       https://doi.org/10.1016/j.envsoft.2012.03.008
    """
    if seed:
        np.random.seed(seed)

    _check_if_num_levels_is_even(num_levels)

    problem = _define_problem_with_groups(problem)

    sample_morris = _sample_morris(problem, N, num_levels)

    if optimal_trajectories:
        if local_optimization and (
            not isinstance(optimal_trajectories, int) or optimal_trajectories > N
        ):
            msg = "optimal_trajectories should be an " f"integer between 2 and {N}"
            raise ValueError(msg)

        sample_morris = _compute_optimised_trajectories(
            problem, sample_morris, N, optimal_trajectories, local_optimization
        )

    sample_morris = scale_samples(sample_morris, problem)

    return sample_morris


def _sample_morris(
    problem: Dict, number_trajectories: int, num_levels: int = 4
) -> np.ndarray:
    """Generate trajectories for groups

    Returns an :math:`N(g+1)`-by-:math:`k` array of `N` trajectories,
    where :math:`g` is the number of groups and :math:`k` is the number
    of factors

    Parameters
    ---------
    problem : dict
        The problem definition
    number_trajectories : int
        The number of trajectories to generate
    num_levels : int, default=4
        The number of grid levels

    Returns
    -------
    numpy.ndarray
    """
    groups = _check_groups(problem)
    group_membership, _ = compute_groups_matrix(groups)
    _check_group_membership(group_membership)

    num_params = group_membership.shape[0]
    num_groups = group_membership.shape[1]

    sample_morris = [
        _generate_trajectory(group_membership, num_levels)
        for _ in range(number_trajectories)
    ]
    sample_morris = np.array(sample_morris)

    return sample_morris.reshape((number_trajectories * (num_groups + 1), num_params))


def _generate_trajectory(
    group_membership: np.ndarray, num_levels: int = 4
) -> np.ndarray:
    """Return a single trajectory

    Return a single trajectory of size :math:`(g+1)`-by-:math:`k`
    where :math:`g` is the number of groups,
    and :math:`k` is the number of factors,
    both implied by the dimensions of `group_membership`

    Parameters
    ---------
    group_membership : np.ndarray
        a k-by-g matrix which notes factor membership of groups
    num_levels : int, default=4
        The number of levels in the grid

    Returns
    -------
    np.ndarray
    """

    delta = _compute_delta(num_levels)

    # Infer number of groups `g` and number of params `k` from
    # `group_membership` matrix
    num_params = group_membership.shape[0]
    num_groups = group_membership.shape[1]

    # Matrix B - size (g + 1) * g -  lower triangular matrix
    B = np.tril(np.ones([num_groups + 1, num_groups], dtype=int), -1)

    P_star = _generate_p_star(num_groups)

    # Matrix J - a (g+1)-by-num_params matrix of ones
    J = np.ones((num_groups + 1, num_params))

    # Matrix D* - num_params-by-num_params matrix which decribes whether
    # factors move up or down
    D_star = np.diag(rd.choice([-1, 1], num_params))

    x_star = _generate_x_star(num_params, num_levels)

    # Matrix B* - size (num_groups + 1) * num_params
    B_star = _compute_b_star(J, x_star, delta, B, group_membership, P_star, D_star)

    return B_star


def _compute_b_star(
    J: np.ndarray,
    x_star: np.ndarray,
    delta: float,
    B: np.ndarray,
    G: np.ndarray,
    P_star: np.ndarray,
    D_star: np.ndarray,
) -> np.ndarray:
    """
    Compute the random sampling matrix B*.

    Parameters
    ----------
    J: matrix of 1's
    x_star: randomly chosen "base value" of x
    delta: parameters variation
    B: sampling matrix (not random)
    G: groups matrix
    P_star: random permutation matrix
    D_star: diagonal matrix with each element being +1 or -1, with equal
            probability

    Returns
    -------
    Random sampling matrix B*
    """
    element_a = J[0, :] * x_star
    element_b = np.matmul(G, P_star).T
    element_c = np.matmul(2.0 * B, element_b)
    element_d = np.matmul((element_c - J), D_star)

    b_star = element_a + (delta / 2.0) * (element_d + J)
    return b_star


def _generate_p_star(num_groups: int) -> np.ndarray:
    """Describe the order in which groups move

    Parameters
    ---------
    num_groups : int

    Returns
    -------
    np.ndarray
        Matrix P* - size (g-by-g)
    """
    p_star = np.eye(num_groups, num_groups)
    rd.shuffle(p_star)
    return p_star


def _generate_x_star(num_params: int, num_levels: int) -> np.ndarray:
    """Generate an 1-by-num_params array to represent initial position for EE

    This should be a randomly generated array in the p level grid
    :math:`\\omega`

    Parameters
    ---------
    num_params : int
        The number of parameters (factors)
    num_levels : int
        The number of levels

    Returns
    -------
    numpy.ndarray
        The initial starting positions of the trajectory
    """
    x_star = np.zeros((1, num_params))
    delta = _compute_delta(num_levels)
    bound = 1 - delta
    grid = np.linspace(0, bound, int(num_levels / 2))

    x_star[0, :] = rd.choice(grid, num_params)

    return x_star


def _compute_optimised_trajectories(
    problem: Dict,
    input_sample: int,
    N: int,
    k_choices: int,
    local_optimization: bool = False,
) -> np.ndarray:
    """
    Calls the procedure to compute the optimum k_choices of trajectories
    from the input_sample.
    If there are groups, then this procedure allocates the groups to the
    correct call here.

    Parameters
    ---------
    problem : dict
        The problem definition
    input_sample :
    N : int
        The number of samples to generate
    k_choices : int
        The number of optimal trajectories
    local_optimization : bool, default=False
        If true, uses local optimisation heuristic
    """
    if local_optimization is False and k_choices > 10:
        msg = "Running optimal trajectories greater than values of 10 \
                will take a long time."
        raise ValueError(msg)

    if np.any((input_sample < 0) | (input_sample > 1)):
        raise ValueError("Input sample must be scaled between 0 and 1")

    num_groups = len(set(problem["groups"]))
    num_params = problem["num_vars"]

    strategy = _choose_optimization_strategy(local_optimization)
    context = SampleMorris(strategy)

    output = context.sample(input_sample, N, num_params, k_choices, num_groups)

    return output


def _check_if_num_levels_is_even(num_levels: int):
    """
    Checks if the number of levels is even. If not, raises a warn.

    Parameters
    ----------
    num_levels: int
        Number of levels
    """
    if not num_levels % 2 == 0:
        warnings.warn("num_levels should be an even number, " "sample may be biased")


def _check_group_membership(group_membership: np.ndarray):
    """
    Checks if the group_membership matrix was correctly defined

    Parameters
    ----------
    group_membership: np.array
        Group membership matrix
    """
    if group_membership is None:
        raise ValueError("Please define the 'group_membership' matrix")
    if not isinstance(group_membership, np.ndarray):
        raise TypeError(
            "Argument 'group_membership' should be formatted \
                         as a numpy np.ndarray"
        )


def _choose_optimization_strategy(local_optimization: bool):
    """
    Choose the strategy to optimize the trajectories.

    Parameters
    ----------
    local_optimization: boolean to indicate if a local optimization should be
                        used.

    Returns
    -------

    """
    if local_optimization:
        # Use local method
        strategy = LocalOptimisation()
    else:
        # Use brute force approach
        strategy = BruteForce()

    return strategy


def cli_parse(parser):
    parser.add_argument(
        "-l",
        "--levels",
        type=int,
        required=False,
        default=4,
        help="Number of grid levels \
                        (Morris only)",
    )
    parser.add_argument(
        "-k",
        "--k-optimal",
        type=int,
        required=False,
        default=None,
        help="Number of optimal trajectories \
                        (Morris only)",
    )
    parser.add_argument(
        "-lo",
        "--local",
        type=bool,
        required=True,
        default=False,
        help="Use the local optimisation method \
                        (Morris with optimization only)",
    )
    return parser


def cli_action(args):
    rd.seed(args.seed)

    problem = read_param_file(args.paramfile)
    param_values = sample(
        problem, args.samples, args.levels, args.k_optimal, args.local
    )

    np.savetxt(
        args.output,
        param_values,
        delimiter=args.delimiter,
        fmt="%." + str(args.precision) + "e",
    )


if __name__ == "__main__":
    common_args.run_cli(cli_parse, cli_action)



import numpy as np


# Non-monotonic Sobol G Function (8 parameters)
# First-order indices:
# x1: 0.7165
# x2: 0.1791
# x3: 0.0237
# x4: 0.0072
# x5-x8: 0.0001
def evaluate(values, a=None, delta=None, alpha=None):
    """Modified Sobol G-function.

    Reverts to original Sobol G-function if delta and alpha are not given.

    .. [1] Saltelli, A., Annoni, P., Azzini, I., Campolongo, F., Ratto, M.,
           Tarantola, S., 2010. Variance based sensitivity analysis of model
           output. Design and estimator for the total sensitivity index.
           Computer Physics Communications 181, 259–270.
           https://doi.org/10.1016/j.cpc.2009.09.018

    Parameters
    ----------
    values : numpy.ndarray
        input variables
    a : numpy.ndarray
        parameter values
    delta : numpy.ndarray
        shift parameters
    alpha : numpy.ndarray
        curvature parameters

    Returns
    -------
    Y : Result of G-function
    """
    if type(values) != np.ndarray:
        raise TypeError("The argument `values` must be a numpy ndarray")

    if a is None:
        a = np.array([0, 1, 4.5, 9, 99, 99, 99, 99])

    if delta is None:
        delta = np.zeros_like(a)
    else:
        if not isinstance(delta, np.ndarray):
            raise TypeError("The argument `delta` must be given as a numpy ndarray")

        delta_inbetween = delta[(delta < 0) | (delta > 1)]
        if delta_inbetween.any():
            raise ValueError(
                "Sobol G function called with delta values less than zero or"
                " greater than one"
            )

    if alpha is None:
        alpha = np.ones_like(a)
    else:
        if not isinstance(alpha, np.ndarray):
            raise TypeError("The argument `alpha` must be given as a numpy ndarray")

        alpha_gto = alpha <= 0.0
        if alpha_gto.any():
            raise ValueError(
                "Sobol G function called with alpha values less than or equal to zero"
            )

    ltz = values < 0
    gto = values > 1

    if ltz.any():
        raise ValueError("Sobol G function called with values less than zero")
    elif gto.any():
        raise ValueError("Sobol G function called with values greater than one")

    Y = np.ones([values.shape[0]])

    for i, row in enumerate(values):
        shift_of_x = row + delta
        integral = np.modf(shift_of_x)[1]
        mod_x = shift_of_x - integral
        temp_y = np.abs(2 * mod_x - 1) ** alpha
        y_elements = ((1 + alpha) * temp_y + a) / (1 + a)
        Y[i] = np.prod(y_elements)

    return Y


def _partial_first_order_variance(a=None, alpha=None):
    if a is None:
        a = [0, 1, 4.5, 9, 99, 99, 99, 99]
    if alpha is None:
        alpha = np.ones_like(a)
    a = np.array(a)

    return np.divide((alpha**2), np.multiply((1 + 2 * alpha), np.square(1 + a)))


def _total_variance(a=None, alpha=None):
    if a is None:
        a = [0, 1, 4.5, 9, 99, 99, 99, 99]
    if alpha is None:
        alpha = np.ones_like(a)

    a = np.array(a)
    return np.add(-1, np.product(1 + _partial_first_order_variance(a, alpha), axis=0))


def sensitivity_index(a, alpha=None):
    a = np.array(a)
    return np.divide(_partial_first_order_variance(a, alpha), _total_variance(a, alpha))


def total_sensitivity_index(a, alpha=None):
    a = np.array(a)

    pv = _partial_first_order_variance(a, alpha)
    tv = _total_variance(a, alpha)
    product_pv = np.product(1 + pv, axis=0)

    return np.divide(pv * np.divide(product_pv, 1 + pv.T), tv)



"""
Created on Dec 20, 2019

@author: @sahin-abdullah

This submodule produces two diffent figures: (1) emulator vs simulator,
(2) regression lines of first order component functions
"""
import matplotlib.pyplot as plt
import numpy as np


def plot(Si):
    # Close all figures
    plt.close("all")

    # Extract necessary data from Si
    try:
        problem = Si.problem
    except AttributeError:
        problem = Si

    Em = Si["Em"]
    RT = Si["RT"]
    Y_em = Si["Y_em"]
    idx = Si["idx"]
    X = Si["X"]
    Y = Si["Y"]

    # Get number of bootstrap from Runtime and sample size N
    K = RT.shape[0]
    N = Y.shape[0]
    row = 2
    col = 5

    try:
        ax = Si._plot()
    except AttributeError:
        # basic bar plot not found or is not available
        pass

    # Plot emulator performance
    Y_p = np.linspace(np.min(Y), np.max(Y), N, endpoint=True)

    start = max(0, K - 10)
    for i in range(start, K):
        # Only showing the last 10
        if (i % (row * col) == 0) or (i == 0):
            fig = plt.figure(frameon=False)
            fig.suptitle("Showing last 10 Bootstrap Trials")
            it = 1

        title_str = f"Trial {K - i}"
        ax = fig.add_subplot(row, col, it, frameon=True, title=title_str)
        ax.plot(Em["Y_e"][:, i], Y[idx[:, i]], "r+", label="Emulator")
        ax.plot(Y_p, Y_p, "darkgray", label="1:1 Line")
        ax.axis("square")
        ax.set_xlim(np.min(Y), np.max(Y))
        ax.set_ylim(np.min(Y), np.max(Y))
        it += 1

        if i == (K - 1):
            handles, labels = ax.get_legend_handles_labels()
            fig.legend(handles, labels, loc="center right", bbox_to_anchor=(1.0, 0.5))

    fig.tight_layout()

    # Now plot regression lines of component functions
    row = 3
    col = 1
    last_bootstrap = idx[:, -1]
    for i in range(problem["num_vars"]):
        if (i % (row * col) == 0) or (i == 0):
            fig = plt.figure(frameon=False)
            it = 1
        title_str = (
            "Regression of parameter " + problem["names"][i] + r"$^{(Last Trial)}$"
        )
        ax = fig.add_subplot(row, col, it, frameon=True, title=title_str)
        ax.plot(X[last_bootstrap, i], Y[last_bootstrap], "r.")
        ax.plot(X[last_bootstrap, i], np.mean(Em["f0"]) + Y_em[:, i], "k.")
        ax.legend(
            [r"$\widetilde{Y}$", "$f_" + str(i + 1) + "$"],
            loc="upper left",
            bbox_to_anchor=(1.04, 1.0),
        )
        it += 1

    fig.tight_layout()
    plt.show()

    if "emulated" in Si:
        emulated = Si["emulated"]
        # Sum of squared residuals
        Y_test = Si["Y_test"]
        ssr = np.sum((emulated - Y_test) ** 2)

        # Plot testing results
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(6, 6))
        ax.plot(emulated, Y_test, "r+", label="Emulator")
        ax.plot(Y_test, Y_test, "darkgray", label="1:1 Line")
        ax.axis("square")
        ax.set_xlabel("Emulator")
        ax.set_ylabel("New Observation")
        ax.set_xlim(np.min(Y_test), np.max(Y_test))
        ax.set_ylim(np.min(Y_test), np.max(Y_test))
        ax.legend(loc="lower right")
        plt.title(f"Testing results\nSSR = {ssr:.2f}")
        fig.tight_layout()
        plt.show()

    return ax


if __name__ == "__main__":
    pass
